"""
Create experimental environment for vivado HLS 
"""

import logging
import typing
from enum import Enum
from pathlib import Path
import shutil

from . import vivado_hls_tcl as vtcl
from . import vivado_report as vrpt
from .tmpdir import TmpDirManager
from .utils import run_script

_define = typing.Union[str, typing.Tuple[str, str]]
_include = str
_define_list = typing.List[_define]
_include_list = typing.List[_include]
_opt_str = typing.Optional[str]


class ExperienceType(Enum):
    CSYNTH_ONLY = 0
    SYNTHESIS = 1
    IMPLEMENTATION = 2


class HDL(Enum):
    VHDL = (0, "vhdl")
    VERILOG = (1, "verilog")

    def __init__(self, *vals):
        self._name = vals[1]

    def get_name(self):
        return self._name


class Experiment:
    def __init__(self,
                 exp_name: str,
                 comp_path: str,
                 comp_name: str,
                 clock_period: int,
                 part: str,
                 standard: str,
                 hdl: HDL,
                 exp_type: ExperienceType,
                 keep_env: bool = False,
                 description: _opt_str = None,
                 version: _opt_str = None,
                 ip_lib: _opt_str = None,
                 vendor: _opt_str = None,
                 logger=None
                 ):
        self._comp_path = comp_path
        self._comp_name = comp_name
        self._clock_period = clock_period
        self._exp_type = exp_type
        self._part = part
        self._standard = standard
        self._includes = []
        self._defines = []
        self._description = description
        self._vendor = vendor
        self._ip_lib = ip_lib
        self._version = version
        self._logger = logger if logger is not None else logging.getLogger("vrs_log")
        self._hdl = hdl
        self._csynth_res = None
        self._syn_impl_res = None
        self._keep_env = keep_env
        self._exp_name = exp_name

    @property
    def clock_period(self):
        return self._clock_period

    @property 
    def name(self):
        return self._exp_name

    def add_includes(self, includes: _include_list):
        self._includes.extend(includes)

    def add_defines(self, defines: _define_list):
        self._defines.extend(defines)

    def _generate_script(self, vivado_script: Path, pipeline: bool, depth_constraint:typing.Optional[int] = None):
        with vivado_script.open('w') as vs:
            vtcl.create_project(vs,
                                "vivado_hls_synthesis",
                                "comp.cpp",
                                self._comp_name,
                                self._includes,
                                self._defines,
                                self._standard,
                                "solution",
                                self._part,
                                self._clock_period,
                                pipeline,
                                depth_constraint
                                )
            vtcl.csynth_solution_script(vs)
            if self._exp_type == ExperienceType.CSYNTH_ONLY:
                return
            vtcl.export_ip_script(
                vs,
                self._exp_type == ExperienceType.IMPLEMENTATION,
                self._hdl.get_name(),
                display_name=self._comp_name,
                descr=self._description,
                version=self._version,
                vendor=self._vendor,
                ip_lib=self._ip_lib
            )

    def run_exp(self, 
                depth_constraint: typing.Optional[int] = None, 
                suffix: str="", 
                before_del = None
               ):
        logger = self._logger
        logger.info(f"Experiment: {self._exp_name}{suffix}")
        comp = Path(self._comp_path).resolve()
        if not comp.exists():
            raise FileNotFoundError(f"Error when starting experiment: component file {comp} does not exist")
        with TmpDirManager(self._exp_name, not self._keep_env):
            comp_copy = Path("comp.cpp")
            shutil.copyfile(comp, comp_copy)

            vivado_script = Path("vivado_hls_script.tcl")
            pipeline = depth_constraint is None or depth_constraint > 1
            self._generate_script(vivado_script, pipeline, depth_constraint)
            run_script(vivado_script)

            vivado_hls_rpt = (Path() / "vivado_hls_synthesis" / "solution" /
                            "syn" / "report" /
                              f"{self._comp_name}_csynth.xml").resolve()

            self._csynth_res = vrpt.parse_syn_report(vivado_hls_rpt)
            if self._exp_type != ExperienceType.CSYNTH_ONLY:
                root = Path() / "vivado_hls_synthesis" / "solution" / "impl" 
                vivado_report = (root / "report" / f"{self._hdl.get_name()}"
                                 / f"{self._comp_name}_export.xml").resolve()
                report_suffix_to_use = ("routed" if self._exp_type ==
                                        ExperienceType.IMPLEMENTATION else
                                        "synth")
                vivado_timing_rpt = (root / self._hdl.get_name() / "report" /
                                     f'{self._comp_name}_timing_{report_suffix_to_use}.rpt').resolve()
                self._syn_impl_res = vrpt.parse_impl_report(vivado_report, vivado_timing_rpt)

            if before_del is not None:
                befor_del()
        return self

    def get_results(self):
        return {"syn": self._csynth_res, "impl": self._syn_impl_res}



"""
Define how to iterate over synthesis
"""

import logging
from pathlib import Path

from . import vivado_exp


class CSVHandler:
    def __init__(self, filename):
        self.file = Path(filename).resolve()
        with self.file.open("w") as f:
            f.write("version, cycles, II, estimated_period, "
                    "real_period, LUT, FF, DSP, BRAM, SRL, Pipelined\n")

    def handle(self, record):
        syn = record["syn"]
        impl = record["impl"]
        with self.file.open('a') as f:
            f.write("{}, {}, {}, {}, {}, {}, {}, {}, {}, {}, {}\n".format(
                syn['Vivado_HLS_Version'],
                syn['worst_case_latency'],
                syn['II'],
                syn['estimated_period'],
                impl['timing'],
                impl['LUT'],
                impl['FF'],
                impl['DSP'],
                impl['BRAM'],
                impl['SRL'],
                syn['pipelined'])
            )


def minimize_latency(exp_settings: vivado_exp.Experiment, handlers):
    logger = logging.getLogger("vrs_log")
    handler = logging.FileHandler("{}.log".format(exp_settings.name), "w")
    handler.setLevel(logging.INFO)
    logger.addHandler(handler)

    logger.info(f"Launching experiment {exp_settings.name}")
    logger.info("Starting without latency constraint")
    ret = exp_settings.run_exp(suffix = "_base").get_results()

    constraint = float(exp_settings.clock_period)

    for h in handlers:
        h.handle(ret)

    timing = float(ret['impl']['timing'])
    cycles = int(ret['syn']['worst_case_latency'])
    old_cycle = cycles + 1
    logger.info(f"Baseline result : {cycles} cycles at a clock period of {timing}ns")
    timings = {cycles: timing}
    up = cycles
    cur_best_timing = timing
    low = -1
    while up > low + 1 and cycles != old_cycle:
        old_cycle = cycles
        if cur_best_timing > constraint:
            mid = 2 * up
            up = 4 * up
        else:
            mid = low + (up - low) // 2
        logger.info(f"Running the experiment with a pipeline depth constraint of {mid}")
        ret = exp_settings.run_exp(
            depth_constraint=mid,
            suffix = f"_{mid}"
        ).get_results()

        for h in handlers:
            h.handle(ret)
        timing = float(ret['impl']['timing'])
        cycles = int(ret['syn']['worst_case_latency'])
        timings[cycles] = timing
        logger.info(f"Result : {cycles} cycles of {timing} ns")
        mid = cycles
        if timing <= constraint:
            up = mid
            cur_best_timing = timing
        else:
            low = mid
        logger.info(f"New interval : [{low},{up}]")
    logger.removeHandler(handler)

    best = up if (low == -1 or timings[low] > constraint) else low
    best_timing = timings[best]

    return best, best_timing



"""
General purpose tcl script creation module
"""
import typing

_opt_str = typing.Optional[str]


def create_project(file, project_name, filename, comp_name, includes, defines,
                   standard, solution_name, part, period, pipeline,
                   depth_constraint):
    """
    Generate a tcl script for an HLS component creation
    :param filename: name of the file containing hls code
    :param period: clock period in ns
    :param part: part code to use
    :param solution_name: name of the solution to create inside the project
    :param file: the file on which the script will be written
    :param project_name: the name of the vivado project to generate
    :param comp_name: the name of the high level component
    :param includes: list of include directories
    :param defines: list of defines
    :param standard: c++ standard code to use
    :param pipeline: should vivado pipeline the design
    :depth_constraint: depth_constraint to set
    """
    file.write("open_project {}\n".format(project_name))
    file.write("set_top {}\n".format(comp_name))
    includes_str = "" if includes is None else " ".join(
        ["-I{}".format(str(s).replace("\\", "/")) for s in includes]
    )
    if defines:
        def format_def(define):
            if len(define) == 2:
                k, v = define
                return f"-D{k}={v}"
            else:
                return f"{define[0]}"

        def_str = " ".join(map(format_def, defines))
    else:
        def_str = ""
    file.write(f'add_files {filename} -cflags "-std={standard} {includes_str} {def_str}"\n')
    file.write('open_solution "{}"\n'.format(solution_name))
    file.write('set_part {{{}}} -tool vivado\n'.format(part))
    file.write('create_clock -period {} -name default\n'.format(
        period
    ))
    file.write('set_clock_uncertainty 0.0 default\n')
    file.write('set_directive_interface -mode ap_ctrl_none [get_top]\n')
    file.write('config_schedule -relax_ii_for_timing=0 -effort high\n')
    file.write('set_directive_inline -recursive [get_top]\n')
    if pipeline and (depth_constraint is None or depth_constraint > 1):
        file.write('set_directive_pipeline -II 1 [get_top]\n')
    if depth_constraint is not None:
        file.write('set_directive_latency -min {} -max {} [get_top]\n'.format(depth_constraint, depth_constraint))


def csynth_solution_script(file):
    """Generate a script in file for synthetizing solution solution_name of project project_name"""
    file.write('csynth_design\n')


def export_ip_script(file,
                     impl: bool,
                     hdl: str,
                     display_name: _opt_str = None,
                     descr: _opt_str = None,
                     version: _opt_str = None,
                     ip_lib: _opt_str = None,
                     vendor: _opt_str = None,
                     ):
    """Generate a script in file for exporting solution solution_name of project project_name"""
    options = [
        "-format ip_catalog",
        f"-rtl {hdl}",
        "-flow {}".format("impl" if impl else "syn")
    ]
    if display_name is not None:
        options.append(f"-display_name {display_name}")
    if descr is not None:
        options.append(f'-description "{descr}"')
    if ip_lib is not None:
        options.append(f'-library {ip_lib}')
    if vendor is not None:
        options.append(f'-vendor {vendor}')
    if version is not None:
        options.append(f'-version {version}')

    command = "export_design {}\n".format(" ".join(options))
    file.write(command)



import setuptools

with open("README.md", "r") as f:
    long_desc = f.read()

setuptools.setup(
    name="vhlsrs", # Replace with your own username
    version="0.0.50",
    author="Luc Forget",
    author_email="luc.forget@insa-lyon.fr",
    description="Package to automate the synthesis of vivado HLS components",
    long_description=long_desc,
    long_description_content_type="text/markdown",
    url="https://gitlab.inria.fr/lforget/vhls_rs",
    packages=setuptools.find_packages(),
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
    install_requires=['plumbum>=1.6'],
)
